@@ -1,14 +1,11 @@
 #!/usr/bin/env ruby -w
 
 begin require 'rubygems'; rescue LoadError; end
-require 'parse_tree'
 require 'sexp_processor'
 require 'unified_ruby'
 
 class Ruby2Ruby < SexpProcessor
-  include UnifiedRuby
-
-  VERSION = '1.1.8'
+  VERSION = '1.2.2'
   LINE_LENGTH = 78
 
   ##
@@ -20,6 +17,7 @@ class Ruby2Ruby < SexpProcessor
                   :flip3,
                   :lasgn,
                   :masgn,
+                  :attrasgn,
                   :op_asgn1,
                   :op_asgn2,
                   :op_asgn_and,
@@ -28,7 +26,18 @@ class Ruby2Ruby < SexpProcessor
                  ]
 
   def self.translate(klass_or_str, method = nil)
-    self.new.process(ParseTree.translate(klass_or_str, method))
+    require 'parse_tree'
+    sexp = ParseTree.translate(klass_or_str, method)
+
+    unifier = Unifier.new
+
+    unifier.processors.each do |p|
+      p.unsupported.delete :cfunc # HACK
+    end
+
+    sexp = unifier.process(sexp)
+
+    self.new.process(sexp)
   end
 
   def initialize
@@ -41,22 +50,25 @@ class Ruby2Ruby < SexpProcessor
     # self.debug[:defn] = /zsuper/
   end
 
-  def process exp
-    exp = Sexp.from_array(exp) if Array === exp unless Sexp === exp
-    super exp
-  end
-
   ############################################################
   # Processors
 
   def process_alias(exp)
-    "alias_method #{process(exp.shift)}, #{process(exp.shift)}"
+    "alias #{process(exp.shift)} #{process(exp.shift)}"
   end
 
   def process_and(exp)
     "(#{process exp.shift} and #{process exp.shift})"
   end
 
+  def process_arglist(exp) # custom made node
+    code = []
+    until exp.empty? do
+      code << process(exp.shift)
+    end
+    code.join ', '
+  end
+
   def process_args(exp)
     args = []
 
@@ -76,8 +88,6 @@ class Ruby2Ruby < SexpProcessor
           args.each_with_index do |name, index|
             args[index] = asgns[name] if asgns.has_key? name
           end
-        when :block_arg then
-          args << "&#{arg.last}"
         else
           raise "unknown arg type #{arg.first.inspect}"
         end
@@ -89,33 +99,6 @@ class Ruby2Ruby < SexpProcessor
     return "(#{args.join ', '})"
   end
 
-  def process_arglist(exp) # custom made node
-    code = []
-    until exp.empty? do
-      code << process(exp.shift)
-    end
-    code.join ', '
-  end
-
-  def process_argscat(exp)
-    args = []
-
-    ary = exp.shift
-    ary.shift # :array
-    until ary.empty? do
-      args << process(ary.shift)
-    end
-
-    args << "*#{process(exp.shift)}"
-    args << process(exp.shift) unless exp.empty? # optional block arg
-
-    args.join ', '
-  end
-
-  def process_argspush(exp)
-    process_arglist(exp)
-  end
-
   def process_array(exp)
     "[#{process_arglist(exp)}]"
   end
@@ -128,13 +111,11 @@ class Ruby2Ruby < SexpProcessor
     case name
     when :[]= then
       rhs = process args.pop
-      args[0] = :arglist if args[0] == :array
       "#{receiver}[#{process(args)}] = #{rhs}"
     else
-      if args then
-        "#{receiver}.#{name.to_s[0..-2]} = #{process(args)[1..-2]}"
-      else
-        "#{receiver}.#{name.to_s[0..-2]}"
+      name = name.to_s.sub(/=$/, '')
+      if args && args != s(:arglist) then
+        "#{receiver}.#{name} = #{process(args)}"
       end
     end
   end
@@ -143,16 +124,16 @@ class Ruby2Ruby < SexpProcessor
     "$#{exp.shift}"
   end
 
+  # TODO: figure out how to do rescue and ensure ENTIRELY w/o begin
   def process_begin(exp)
-    is_rescue = exp.first.first == :rescue rescue false
     code = []
     code << "begin"
     until exp.empty?
       src = process(exp.shift)
-      src = indent(src) unless src =~ /(^|\n)rescue/ # ensures no level 0 rescues
+      src = indent(src) unless src =~ /(^|\n)rescue/ # ensure no level 0 rescues
       code << src
     end
-    code << "end" unless is_rescue
+    code << "end"
     return code.join("\n")
   end
 
@@ -181,30 +162,10 @@ class Ruby2Ruby < SexpProcessor
     return result
   end
 
-  def process_block_arg(exp)
-    "&#{exp.shift}"
-  end
+  def process_block_pass exp
+    raise "huh?: #{exp.inspect}" if exp.size > 1
 
-  def process_block_pass(exp)
-    bname = s(:block_arg, process(exp.shift)) # FIX
-    call = exp.shift
-
-    if Array === call.last then # HACK - I _really_ need rewrites to happen first
-      case call.last.first
-      when :splat then
-        call << [:array, call.pop]
-      when :array then
-        # do nothing
-      else
-        has_args = Array === call.last and call.last.first == :array
-        call << [:array] unless has_args
-      end
-      call.last << bname
-    else
-      call << [:array, bname]
-    end
-
-    process(call)
+    "&#{process exp.shift}"
   end
 
   def process_break(exp)
@@ -225,29 +186,29 @@ class Ruby2Ruby < SexpProcessor
       Ruby2Ruby::ASSIGN_NODES.include? receiver_node_type
 
     name = exp.shift
-    args_exp = exp.shift rescue nil
-    if args_exp && args_exp.first == :array # FIX
-      args = "#{process(args_exp)[1..-2]}"
-    else
-      args = process args_exp
-      args = nil if args.empty?
-    end
+    args = exp.shift rescue nil
 
     case name
     when :<=>, :==, :<, :>, :<=, :>=, :-, :+, :*, :/, :%, :<<, :>>, :** then
-      "(#{receiver} #{name} #{args})"
+      "(#{receiver} #{name} #{process args})"
     when :[] then
-      "#{receiver}[#{args}]"
+      receiver = "self" if receiver.nil?
+      "#{receiver}[#{process args}]"
+    when :[]= then
+      receiver = "self" if receiver.nil?
+      lhs = args.pop
+      "#{receiver}[#{process args}] = #{process lhs}"
     when :"-@" then
       "-#{receiver}"
     when :"+@" then
       "+#{receiver}"
     else
-      unless receiver.nil? then
-        "#{receiver}.#{name}#{args ? "(#{args})" : args}"
-      else
-        "#{name}#{args ? "(#{args})" : args}"
-      end
+      args     = process args
+      args     = nil            if args.empty?
+      args     = "(#{args})"    if args
+      receiver = "#{receiver}." if receiver
+
+      "#{receiver}#{name}#{args}"
     end
   end
 
@@ -274,7 +235,14 @@ class Ruby2Ruby < SexpProcessor
   end
 
   def process_cdecl(exp)
-    "#{exp.shift} = #{process(exp.shift)}"
+    lhs = exp.shift
+    lhs = process lhs if Sexp === lhs
+    unless exp.empty? then
+      rhs = process(exp.shift)
+      "#{lhs} = #{rhs}"
+    else
+      lhs.to_s
+    end
   end
 
   def process_class(exp)
@@ -305,34 +273,6 @@ class Ruby2Ruby < SexpProcessor
     "#{exp.shift} = #{process(exp.shift)}"
   end
 
-  # (a, lit1)      => "a = 1"
-  # (a, (b, lit2)) => "a = b = 2"
-  # (a, (b))       => ""
-
-  def process_dasgn_curr(exp)
-    lhs = exp.shift.to_s
-    rhs = (exp.empty? ? nil : exp.shift)
-    if rhs.nil? then
-      if self.context[1] == :block then
-        return ''
-      end
-
-      return lhs
-    end
-    return "#{lhs} = #{process rhs}" unless rhs.first == :dasgn_curr
-
-    # keep recursing. ensure that the leaf node assigns to _something_
-    "#{lhs} = #{process rhs}"
-  end
-
-  def process_dasgn(exp)
-    if exp.size == 1 then
-      exp.shift.to_s
-    else
-      "#{exp.shift} = #{process(exp.shift)}"
-    end
-  end
-
   def process_defined(exp)
     "defined? #{process(exp.shift)}"
   end
@@ -368,7 +308,14 @@ class Ruby2Ruby < SexpProcessor
   end
 
   def process_defs(exp)
-    exp.unshift "#{process(exp.shift)}.#{exp.shift}"
+    lhs  = exp.shift
+    var = [:self, :cvar, :dvar, :ivar, :gvar, :lvar].include? lhs.first
+    name = exp.shift
+
+    lhs = process(lhs)
+    lhs = "(#{lhs})" unless var
+
+    exp.unshift "#{lhs}.#{name}"
     process_defn(exp)
   end
 
@@ -396,41 +343,29 @@ class Ruby2Ruby < SexpProcessor
     ":#{process_dstr(exp)}"
   end
 
-  def process_dvar(exp)
-    exp.shift.to_s
-  end
-
   def process_dxstr(exp)
     "`#{process_dstr(exp)[1..-2]}`"
   end
 
   def process_ensure(exp)
     body = process exp.shift
-    ens  = process exp.shift
+    ens  = exp.shift
+    ens  = nil if ens == s(:nil)
+    ens  = process(ens) || "# do nothing"
+
+    body.sub!(/\n\s*end\z/, '')
+
     return "#{body}\nensure\n#{indent ens}"
   end
 
   def process_evstr(exp)
-    process exp.shift
+    exp.empty? ? '' : process(exp.shift)
   end
 
   def process_false(exp)
     "false"
   end
 
-  # TODO: remove for unified
-  def process_fcall(exp)
-    recv = exp.shift unless Symbol === exp.first # HACK conditional - some not getting rewritten?
-    name = exp.shift.to_s
-    args = exp.shift
-    code = []
-    unless args.nil? then
-      args[0] = :arglist if args.first == :array
-      code << process(args)
-    end
-    return code.empty? ? name : "#{name}(#{code.join(', ')})"
-  end
-
   def process_flip2(exp)
     "#{process(exp.shift)}..#{process(exp.shift)}"
   end
@@ -462,12 +397,22 @@ class Ruby2Ruby < SexpProcessor
   def process_hash(exp)
     result = []
     until exp.empty?
-      result << "#{process(exp.shift)} => #{process(exp.shift)}"
+      lhs = process(exp.shift)
+      rhs = exp.shift
+      t = rhs.first
+      rhs = process rhs
+      rhs = "(#{rhs})" unless [:lit, :str].include? t # TODO: verify better!
+
+      result << "#{lhs} => #{rhs}"
     end
 
     case self.context[1]
     when :arglist, :argscat then
-      return "#{result.join(', ')}" # HACK - this will break w/ 2 hashes as args
+      unless result.empty? then
+        return "#{result.join(', ')}" # HACK - this will break w/ 2 hashes as args
+      else
+        return "{}"
+      end
     else
       return "{ #{result.join(', ')} }"
     end
@@ -546,8 +491,10 @@ class Ruby2Ruby < SexpProcessor
     result << "#{iter} #{b}"
     result << " |#{args}|" if args
     result << "\n"
-    result << indent(body.strip)
-    result << "\n"
+    if body then
+      result << indent(body.strip)
+      result << "\n"
+    end
     result << e
     result.join
   end
@@ -584,12 +531,6 @@ class Ruby2Ruby < SexpProcessor
     lhs = exp.shift
     rhs = exp.empty? ? nil : exp.shift
 
-    unless exp.empty? then
-      rhs[-1] = splat(rhs[-1]) unless rhs == s(:splat)
-      lhs << rhs
-      rhs = exp.shift
-    end
-
     case lhs.first
     when :array then
       lhs.shift
@@ -601,7 +542,7 @@ class Ruby2Ruby < SexpProcessor
           process(l)
         end
       end
-    when :dasgn_curr then
+    when :lasgn then
       lhs = [ splat(lhs.last) ]
     when :splat then
       lhs = [ :"*" ]
@@ -610,25 +551,18 @@ class Ruby2Ruby < SexpProcessor
     end
 
     if context[1] == :iter and rhs then
-      lhs << splat(rhs.last)
+      lhs << splat(rhs[1])
       rhs = nil
     end
 
     unless rhs.nil? then
       t = rhs.first
-      rhs = if t == :argscat then
-              rhs.shift
-              process_argscat(rhs)
-            else
-              r = process(rhs)
-              r = r[1..-2] if t != :to_ary
-              r
-            end
+      rhs = process rhs
+      rhs = rhs[1..-2] if t == :array # FIX: bad? I dunno
       return "#{lhs.join(", ")} = #{rhs}"
     else
       return lhs.join(", ")
     end
-
   end
 
   def process_match(exp)
@@ -673,13 +607,13 @@ class Ruby2Ruby < SexpProcessor
   end
 
   def process_op_asgn1(exp)
-    # [[:lvar, :b], [:array, [:lit, 1]], :"||", [:lit, 10]]
+    # [[:lvar, :b], [:arglist, [:lit, 1]], :"||", [:lit, 10]]
     lhs = process(exp.shift)
     index = process(exp.shift)
     msg = exp.shift
     rhs = process(exp.shift)
 
-    "#{lhs}#{index} #{msg}= #{rhs}"
+    "#{lhs}[#{index}] #{msg}= #{rhs}"
   end
 
   def process_op_asgn2(exp)
@@ -693,18 +627,18 @@ class Ruby2Ruby < SexpProcessor
     "#{lhs}.#{index} #{msg}= #{rhs}"
   end
 
-  def process_op_asgn_or(exp)
-    # a ||= 1
-    # [[:lvar, :a], [:lasgn, :a, [:lit, 1]]]
-    exp.shift
-    process(exp.shift).sub(/=/, '||=')
-  end
-
   def process_op_asgn_and(exp)
     # a &&= 1
     # [[:lvar, :a], [:lasgn, :a, [:lit, 1]]]
     exp.shift
-    process(exp.shift).sub(/=/, '&&=')
+    process(exp.shift).sub(/\=/, '&&=')
+  end
+
+  def process_op_asgn_or(exp)
+    # a ||= 1
+    # [[:lvar, :a], [:lasgn, :a, [:lit, 1]]]
+    exp.shift
+    process(exp.shift).sub(/\=/, '||=')
   end
 
   def process_or(exp)
@@ -719,97 +653,40 @@ class Ruby2Ruby < SexpProcessor
     "redo"
   end
 
-  def process_resbody(exp) # TODO: rewrite this fucker
-    code = []
+  def process_resbody exp
+    args = exp.shift
+    body = process(exp.shift) || "# do nothing"
 
-    sexp = exp
-    until exp.empty? and (sexp.nil? or sexp.empty?)
-      list = sexp.shift
-      body = sexp.shift
+    name =   args.lasgn true
+    name ||= args.iasgn true
+    args = process(args)[1..-2]
+    args = " #{args}" unless args.empty?
+    args += " => #{name[1]}" if name
 
-      var = if list and
-                list.size > 1 and
-                [:lasgn, :dasgn, :dasgn_curr].include? list.last.first then
-              list.pop[1]
-            else
-              nil
-            end
-
-      # FIX: omg this is horrid. I should be punished
-      var = body.delete_at(1)[1] if
-        [:dasgn_curr, :dasgn].include? body[1][0] unless
-        var or body.nil? rescue nil
-
-      if list and list.size > 1 then
-        list[0] = :arglist
-        code << "rescue #{process(list)}"
-      else
-        code << "rescue"
-      end
-
-      code.last << " => #{var}" if var
-
-      if body then
-        code << indent(process(body)).chomp
-      else
-        code << indent("# do nothing")
-      end
-
-      unless exp.empty? then
-        sexp = exp.shift
-        assert_type sexp, :resbody
-        sexp.shift
-      end
-    end
-
-    code.join("\n")
+    "rescue#{args}\n#{indent body}"
   end
 
-  def process_rescue(exp)
-    # TODO: rewrite this
-    #
-    # a = b rescue c            =>                [lasgn a [rescue b c]]
-    # begin; a = b; rescue c    => [begin [rescue [lasgn a b] c]]
+  def process_rescue exp
+    body = process(exp.shift) unless exp.first.first == :resbody
+    els  = process(exp.pop)   unless exp.last.first  == :resbody
 
-    current = self.context[1]
-    case current
-    when :begin, :ensure, :block then
-      body = (exp.first.first == :resbody) ? nil : process(exp.shift)
-      resbody = exp.empty? ? '' : process(exp.shift)
-      els = exp.empty? ? nil : process(exp.shift)
+    body ||= "# do nothing"
+    simple = exp.size == 1
 
-      code = []
-      code << indent(body) if body
-      code << resbody
-      if els then
-        code << "else"
-        code << indent(els)
-      else
-        unless [:block].include? current then
-          code << "end\n" unless current == :ensure
-        else
-          r = [body, resbody.gsub(/rescue\n\s+/, 'rescue ')].compact.join(' ')
-          code = [r] if (@indent+r).size < LINE_LENGTH and r !~ /\n/
-        end
-      end
-
-      code.join("\n").chomp
-    else # a rescue b and others
-      body = process exp.shift
-      assert_type exp.first, :resbody
+    resbodies = []
+    until exp.empty? do
       resbody = exp.shift
-      resbody.shift # resbody
-      resbody.shift # nil (no types for expression form)
-      resbody = resbody.shift # actual code
+      simple &&= resbody[1] == s(:array) && resbody[2] != nil
+      resbodies << process(resbody)
+    end
 
-      resbody = process resbody
-      code = "#{body} rescue #{resbody}"
-      case current
-      when :hash then # HACK move to process_hash
-        "(#{code})"
-      else
-        code
-      end
+    if els then
+      "#{indent body}\n#{resbodies.join("\n")}\nelse\n#{indent els}"
+    elsif simple then
+      resbody = resbodies.first.sub(/\n\s*/, ' ')
+      "#{body} #{resbody}"
+    else
+      "#{indent body}\n#{resbodies.join("\n")}"
     end
   end
 
@@ -852,13 +729,20 @@ class Ruby2Ruby < SexpProcessor
   end
 
   def process_super(exp)
-    args = exp.shift
-    args[0] = :arglist if args[0] == :array
-    "super(#{process(args)})"
+    args = []
+    until exp.empty? do
+      args << process(exp.shift)
+    end
+
+    "super(#{args.join(', ')})"
   end
 
   def process_svalue(exp)
-    process(exp.shift)
+    code = []
+    until exp.empty? do
+      code << process(exp.shift)
+    end
+    code.join(", ")
   end
 
   def process_to_ary(exp)
@@ -909,24 +793,18 @@ class Ruby2Ruby < SexpProcessor
   end
 
   def process_yield(exp)
-    args = exp.empty? ? nil : exp.shift
-    if args then
-      args[0] = :arglist if args.first == :array
-      args = process(args)
+    args = []
+    until exp.empty? do
+      args << process(exp.shift)
     end
 
-    # "yield" + (args ? "(#{args})" : "")
-    if args then
-      "yield(#{args})"
+    unless args.empty? then
+      "yield(#{args.join(', ')})"
     else
       "yield"
     end
   end
 
-  def process_zarray(exp)
-    "[]"
-  end
-
   def process_zsuper(exp)
     "super"
   end
@@ -954,24 +832,52 @@ class Ruby2Ruby < SexpProcessor
   ############################################################
   # Rewriters:
 
-  def rewrite_rescue exp
-    exp = s(:begin, exp) if
-      context[1] == :block unless
-      context[2] == :scope and [:defn, :defs].include? context[3]
+  def rewrite_attrasgn exp
+    if context.first(2) == [:array, :masgn] then
+      exp[0] = :call
+      exp[2] = exp[2].to_s.sub(/=$/, '').to_sym
+    end
+
     exp
   end
 
+  def rewrite_ensure exp
+    exp = s(:begin, exp) unless context.first == :begin
+    exp
+  end
+
+  def rewrite_rescue exp
+    complex = false
+    complex ||= exp.size > 3
+    complex ||= exp.block
+    complex ||= exp.find_nodes(:resbody).any? { |n| n.array != s(:array) }
+    complex ||= exp.find_nodes(:resbody).any? { |n| n.last.nil? }
+
+    handled = context.first == :ensure
+
+    exp = s(:begin, exp) if complex unless handled
+
+    exp
+  end
+
+  def rewrite_svalue(exp)
+    case exp.last.first
+    when :array
+      s(:svalue, *exp[1][1..-1])
+    when :splat
+      exp
+    else
+      raise "huh: #{exp.inspect}"
+    end
+  end
+
   ############################################################
   # Utility Methods:
 
   def util_dthing(exp, regx = false)
     s = []
     suck = true
-    if suck then
-      x = exp.shift.gsub(/"/, '\"').gsub(/\n/, '\n')
-    else
-      x = exp.shift.dump[1..-2]
-    end
+    x = exp.shift.gsub(/"/, '\"').gsub(/\n/, '\n')
     x.gsub!(/\//, '\/') if regx
 
     s << x
@@ -981,18 +887,15 @@ class Ruby2Ruby < SexpProcessor
       when Sexp then
         case pt.first
         when :str then
-          if suck then
-            x = pt.last.gsub(/"/, '\"').gsub(/\n/, '\n')
-          else
-            x = pt.last.dump[1..-2]
-          end
+          x = pt.last.gsub(/"/, '\"').gsub(/\n/, '\n')
           x.gsub!(/\//, '\/') if regx
           s << x
         else
           s << '#{' << process(pt) << '}' # do not use interpolation here
         end
       else
-        # do nothing - yet
+        # HACK: raise "huh?: #{pt.inspect}"
+        # do nothing for now
       end
     end
 
@@ -1000,95 +903,38 @@ class Ruby2Ruby < SexpProcessor
   end
 
   def util_module_or_class(exp, is_class=false)
-    s = "#{exp.shift}"
+    result = []
+
+    name = exp.shift
+    name = process name if Sexp === name
+
+    result << name
 
     if is_class then
       superk = process(exp.shift)
-      s << " < #{superk}" if superk
+      result << " < #{superk}" if superk
     end
 
-    s << "\n"
+    result << "\n"
 
     body = []
     begin
-      code = process(exp.shift).chomp
-      body << code unless code.nil? or code.empty?
+      code = process(exp.shift)
+      body << code.chomp unless code.nil? or code.chomp.empty?
     end until exp.empty?
+
     unless body.empty? then
       body = indent(body.join("\n\n")) + "\n"
     else
       body = ""
     end
-    s + body + "end"
+    result << body
+    result << "end"
+
+    result.join
   end
 
   def indent(s)
-    s.to_s.map{|line| @indent + line}.join
-  end
-end
-
-RubyToRuby = Ruby2Ruby # For backwards compatibilty... TODO: remove 2008-03-28
-
-class Method
-  def with_class_and_method_name
-    if self.inspect =~ /<Method: (.*)\#(.*)>/ then
-      klass = eval $1
-      method  = $2.intern
-      raise "Couldn't determine class from #{self.inspect}" if klass.nil?
-      return yield(klass, method)
-    else
-      raise "Can't parse signature: #{self.inspect}"
-    end
-  end
-
-  def to_sexp
-    with_class_and_method_name do |klass, method|
-      ParseTree.new(false).parse_tree_for_method(klass, method)
-    end
-  end
-
-  def to_ruby
-    Ruby2Ruby.new.process(self.to_sexp)
-  end
-end
-
-class ProcStoreTmp
-  @@n = 0
-  def self.name
-    @@n += 1
-    return :"myproc#{@@n}"
-  end
-end
-
-class UnboundMethod
-  def to_ruby
-    name = ProcStoreTmp.name
-    ProcStoreTmp.send(:define_method, name, self)
-    m = ProcStoreTmp.new.method(name)
-    result = m.to_ruby.sub(/def #{name}(?:\(([^\)]*)\))?/,
-                           'proc { |\1|').sub(/end\Z/, '}')
-    return result
-  end
-end
-
-class Proc
-  def to_method
-    name = ProcStoreTmp.name
-    ProcStoreTmp.send(:define_method, name, self)
-    ProcStoreTmp.new.method(name)
-  end
-
-  def to_sexp
-    body = self.to_method.to_sexp[2][1..-1]
-    [:proc, *body]
-  end
-
-  def to_ruby
-    ruby = self.to_method.to_ruby
-    ruby.sub!(/\A(def \S+)\(([^\)]*)\)/, '\1 |\2|')     # move args
-    ruby.sub!(/\Adef[^\n\|]+/, 'proc { ')               # strip def name
-    ruby.sub!(/end\Z/, '}')                             # strip end
-    ruby.gsub!(/\s+$/, '')                              # trailing WS bugs me
-    ruby
+    s.to_s.split(/\n/).map{|line| @indent + line}.join("\n")
   end
 end
